from django.core.exceptions import ImproperlyConfigured
from django.utils import timezone
from django.conf import settings
from django.shortcuts import render

from drf_yasg.utils import swagger_auto_schema

from rest_framework.response import Response
from rest_framework import viewsets, status

from bcmr.auth import HeaderAuthentication
from bcmr.utils import generate_auth_token
from bcmr.serializers import *
from bcmr.filters import *
from bcmr.models import *
from bcmr.forms import *


def get_or_create_owner(auth_token_id):
    if auth_token_id:
        return AuthToken.objects.get(id=auth_token_id)
    else:
        return generate_auth_token()


class TokenViewSet(viewsets.ModelViewSet):
    queryset = Token.objects.all()
    filterset_class = TokenFilter
    filter_backends = (filters.DjangoFilterBackend, )
    authentication_classes = (HeaderAuthentication, )
    serializer_class = EmptySerializer
    serializer_classes = {
        'create': CashTokenSerializer,
        'list': RegistryIdentitySerializer,
        'update': CashTokenSerializer,
        'partial_update': CashTokenSerializer,
        'retrieve': RegistryIdentitySerializer
    }

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)            
        serializer.validated_data['owner'] = get_or_create_owner(request.META.get(settings.AUTH_HEADER))
        self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)

    def get_serializer_class(self):
        if not isinstance(self.serializer_classes, dict):
            raise ImproperlyConfigured("serializer_classes should be a dict mapping.")

        if self.action in self.serializer_classes.keys():
            return self.serializer_classes[self.action]
        return super().get_serializer_class()


class RegistryViewSet(viewsets.ModelViewSet):
    queryset = Registry.objects.all()
    filterset_class = RegistryFilter
    filter_backends = (filters.DjangoFilterBackend, )
    authentication_classes = (HeaderAuthentication, )
    serializer_class = EmptySerializer
    serializer_classes = {
        'create': RegistrySerializer,
        'list': NoOwnerRegistrySerializer,
        'update': RegistrySerializer,
        'partial_update': RegistrySerializer,
        'retrieve': BcmrRegistrySerializer
    }

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)            
        serializer.validated_data['owner'] = get_or_create_owner(request.META.get(settings.AUTH_HEADER))
        self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)
    
    @swagger_auto_schema(responses={200: BcmrRegistrySerializer})
    def update(self, request, *args, **kwargs):
        partial = kwargs.pop('partial', False)
        instance = self.get_object()
        instance.latest_revision = timezone.now()
        instance.save()

        serializer = self.get_serializer(instance, data=request.data, partial=partial)
        serializer.is_valid(raise_exception=True)
        self.perform_update(serializer)

        bcmr = BcmrRegistrySerializer(instance)
        return Response(bcmr.data)

    def get_serializer_class(self):
        if not isinstance(self.serializer_classes, dict):
            raise ImproperlyConfigured("serializer_classes should be a dict mapping.")

        if self.action in self.serializer_classes.keys():
            return self.serializer_classes[self.action]
        return super().get_serializer_class()

 
def add_token(request):
    submitted = False
    message = ''
    owner = ''

    if request.method == 'POST':
        queryDict = request.POST
        data = dict(queryDict.lists())

        category = data['category'][0]
        name = data['name'][0]
        description = data['description'][0]
        symbol = data['symbol'][0]
        decimals = int(data['decimals'][0])
        auth_token = data['auth_token'][0]
        icon = data['icon'][0]
        
        if Token.objects.filter(category=category).exists():
            message = 'Token category already exists!'
        else:
            try:
                owner = get_or_create_owner(auth_token)
                token = Token(
                    category=category,
                    name=name,
                    description=description,
                    symbol=symbol,
                    decimals=decimals,
                    owner=owner,
                    icon=icon
                )
                token.save()
                message = 'Token added!'
            except:
                message = 'Invalid auth token!'

        submitted = True

    context = {
        'form': TokenForm(),
        'submitted': submitted,
        'message': message,
        'owner': owner if type(owner) is str else owner.id
    }
    return render(request, 'bcmr/add_token.html', context)



# Generated by Django 3.1.6 on 2023-02-09 08:04

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Token',
            fields=[
                ('category', models.CharField(max_length=255, primary_key=True, serialize=False, unique=True)),
                ('name', models.CharField(max_length=255)),
                ('description', models.CharField(blank=True, default='', max_length=255, null=True)),
                ('symbol', models.CharField(max_length=100)),
                ('decimals', models.PositiveIntegerField(default=0)),
                ('icon', models.ImageField(blank=True, null=True, upload_to='')),
                ('date_created', models.DateTimeField(default=django.utils.timezone.now)),
                ('status', models.CharField(choices=[('active', 'Active'), ('inactive', 'Inactive'), ('burned', 'Burned')], default='active', max_length=10)),
            ],
            options={
                'ordering': ('name',),
            },
        ),
        migrations.CreateModel(
            name='Registry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('major', models.PositiveIntegerField(default=0)),
                ('minor', models.PositiveIntegerField(default=0)),
                ('patch', models.PositiveIntegerField(default=0)),
                ('name', models.CharField(max_length=255)),
                ('description', models.CharField(blank=True, max_length=255, null=True)),
                ('date_created', models.DateTimeField(default=django.utils.timezone.now)),
                ('latest_revision', models.DateTimeField(default=django.utils.timezone.now)),
                ('tokens', models.ManyToManyField(to='bcmr.Token')),
            ],
            options={
                'verbose_name_plural': 'Registries',
            },
        ),
    ]



from rest_framework import serializers

from django.conf import settings

from bcmr.models import *


class EmptySerializer(serializers.Serializer):
    pass


class CashTokenSerializer(serializers.ModelSerializer):
    class Meta:
        model = Token
        fields = (
            'category',
            'name',
            'symbol',
            'description',
            'decimals',
            'icon',
            'status',
            'date_created',
            'owner',
        )
        read_only_fields = (
            'date_created',
            'owner',
        )


class RegistryIdentitySerializer(serializers.ModelSerializer):
    time = serializers.SerializerMethodField()
    token = serializers.SerializerMethodField()
    uris = serializers.SerializerMethodField()

    class Meta:
        model = Token
        fields = (
            'name',
            'description',
            'time',
            'token',
            'status',
            'uris',
        )

    def get_time(self, obj):
        return {
            'begin': obj.date_created
        }

    def get_token(self, obj):
        return {
            'category': obj.category,
            'symbol': obj.symbol,
            'decimals': obj.decimals
        }

    def get_uris(self, obj):
        if obj.icon:
            return {
                'icon': f'{settings.DOMAIN}{obj.icon.url}'
            }
        return {}


class RegistrySerializer(serializers.ModelSerializer):
    class Meta:
        model = Registry
        fields = (
            'id',
            'name',
            'description',
            'tokens',
            'owner',
        )
        read_only_fields = (
            'owner',
            'id',
        )


class NoOwnerRegistrySerializer(serializers.ModelSerializer):
    class Meta:
        model = Registry
        fields = (
            'id',
            'name',
            'description',
            'tokens',
        )
        read_only_fields = (
            'id',
        )


class BcmrRegistrySerializer(serializers.ModelSerializer):
    version = serializers.SerializerMethodField()
    latestRevision = serializers.SerializerMethodField()
    registryIdentity = serializers.SerializerMethodField()
    identities = serializers.SerializerMethodField()

    class Meta:
        model = Registry
        fields = (
            'version',
            'latestRevision',
            'registryIdentity',
            'identities',
        )

    def get_version(self, obj):
        return {
            'major': obj.major,
            'minor': obj.minor,
            'patch': obj.patch
        }

    def get_latestRevision(self, obj):
        return obj.latest_revision

    def get_registryIdentity(self, obj):
        return {
            'name': obj.name,
            'description': obj.description,
            'time': {
                'begin': obj.date_created
            }
        }

    def get_identities(self, obj):
        tokens = obj.tokens.all()
        identities = {}

        for token in tokens:
            t = RegistryIdentitySerializer(token).data
            identities[t['token']['category']] = [t]

        return identities
